/*
// File:		ImapResponseHandler.cpp
//
// Version:		mail.it library 1.0 sample file
//
// Copyright:	(C) 2007 Dacons LLP.
//
// Bugs:		If you have problems or suggestions, please contact
//				the Online Support at: http://www.dacons.net/support
*/

#include PCH_FILE_NAME
#include "ImapResponseHandler.hpp"
#include "ImapFolder.hpp"
#include <iostream>
#include "ImapMessageData.hpp"
#include "StreamUtils.hpp"
#include <assert.h>

ImapResponseHandler::ImapResponseHandler()
{
}

ImapResponseHandler::~ImapResponseHandler()
{
	ClearFolderList();
	ClearSubscribedFolderList();
}

void
ImapResponseHandler::ClearList(
		ImapFolderList& inList
	)
{
	ImapFolderList::iterator iter = inList.begin();
	while (iter != inList.end())
	{
		delete (*iter);
		++iter;
	}
	inList.clear();
}

void
ImapResponseHandler::ClearFolderList()
{
	ClearList(mFolders);
}

void
ImapResponseHandler::ClearSubscribedFolderList()
{
	ClearList(mSubscribedFolders);
}

/* "Inbox" mailbox is case insensitive*/
bool
ImapResponseHandler::IsInbox(
		EmaiConstUniCharPtr inName
	)
{
	EmaiUniChar inboxName[] = {'i', 'n', 'b', 'o', 'x', '\0'};
	EmaiConstUniCharPtr p1 = inName;
	EmaiConstUniCharPtr p2 = inboxName;
	int c1, c2;
	
	while ((c1 = tolower(*p1++)) == (c2 = tolower(*p2++)))
		if (c1 == '\0')
			return true;

	return false;
}

#if 0
static inline bool
IsInboxInfefior(
		EmaiConstUniCharPtr inName
	)
{
	EmaiUniChar inboxName[] = {'i', 'n', 'b', 'o', 'x', '\0', '\0'};
	EmaiConstUniCharPtr p1 = inName;
	EmaiConstUniCharPtr p2 = inboxName;
	int c1, c2;
	
	while ((c1 = tolower(*p1++)) == (c2 = tolower(*p2++)))
		if (c1 == '\0')
			return true;

	return false;
}
#endif

ImapFolder*
ImapResponseHandler::FindFolder(
		ImapFolderList& inList,
		EmaiConstUniCharPtr inName
	)
{
	ImapFolderList::iterator iter = inList.begin();
	while (iter != inList.end())
	{
		if ((*iter)->GetName() == inName)
			return *iter;
		
		if (IsInbox(inName) && IsInbox((*iter)->GetName().c_str()))
			return *iter;
		++iter;
	}
	return NULL;
}

ImapFolder*
ImapResponseHandler::FindFolderByName(
		EmaiConstUniCharPtr inName
	)
{
	ImapFolder* folder = FindFolder(mFolders, inName);
	return folder;
}

ImapFolder*
ImapResponseHandler::FindSubscribedFolderByName(
		EmaiConstUniCharPtr inName
	)
{
	return FindFolder(mSubscribedFolders, inName);
}

void
ImapResponseHandler::DeleteFolderFromFolderList(
		EmaiConstUniCharPtr inName
	)
{
	ImapFolderList::iterator iter = mFolders.begin();
	while (iter != mFolders.end())
	{
		if (((*iter)->GetName() == inName) ||
			(IsInbox(inName) && IsInbox((*iter)->GetName().c_str())))
		{
			delete (*iter);
			mFolders.erase(iter);
			return;
		}
		++iter;
	}
}

void
ImapResponseHandler::RenameFoldersInFolderList(
		EmaiConstUniCharPtr inName,
		EmaiConstUniCharPtr inNewName
	)
{
	if (IsInbox(inName))
	{
		//Renaming INBOX is permitted, and has special behavior.  It moves 
		//all messages in INBOX to a new mailbox with the given name, 
		//leaving INBOX empty.  If the server implementation supports 
		//inferior hierarchical names of INBOX, these are unaffected by a 
		//rename of INBOX. 
		ImapFolder* folder = FindFolderByName(inName);
		if (folder != NULL)
			folder->SetName(inNewName);
	
		//TODO: add object for mailbox
	}
	else
	{
		//If the name has inferior hierarchical names, then the inferior 
		//hierarchical names MUST also be renamed.  For example, a rename of 
		//"foo" to "zap" will rename "foo/bar" (assuming "/" is the 
		//hierarchy delimiter character) to "zap/bar".
		
		ImapFolderList::iterator iter = mFolders.begin();
		while (iter != mFolders.end())
		{
			Emai::UniString name(inName);
			name.append(1, (*iter)->GetDelimiter());
			
			const Emai::UniString& folderName = (*iter)->GetName();
			if (folderName == inName)
				(*iter)->SetName(inNewName);
			else if (folderName.find(name) == 0)
			{
				Emai::UniString newName = folderName;
				newName.replace(0, name.length() - 1, inNewName);
				(*iter)->SetName(newName);
			}
			++iter;
		}
	}
}

void
ImapResponseHandler::ProcessStatus(
		const EmaiImapCommandData* inCommandData,
		const EmaiStatusResponse& inResponse
	)
{
	EmaiUint32 responseType = inResponse.responseType & ~EmaiTaggedStatusResponse;
	if ((inResponse.responseType & EmaiTaggedStatusResponse) != 0)
	{
		if (inResponse.responseType != EmaiImapTaggedOkResponse)
			std::cout << "!!!Command failed: " << inResponse.responseText.text
									<< std::endl;
	}
	
	ImapFolder* folder = NULL;
	if (inCommandData->sourceMailboxName != NULL)
		folder = FindFolderByName(inCommandData->sourceMailboxName);

	switch (responseType)
	{
		case EmaiImapOkResponse:
		{
			switch (inResponse.responseText.code)
			{
				case EmaiPermanentFlagsCode:
					assert(inCommandData->sourceMailboxName != NULL);
					assert(folder != NULL);
					folder->SetFlags(inResponse.responseText.param.permanentFlags.flags,
										Emai::StringList(inResponse.responseText.param.permanentFlags.customFlags, true));
					break;
				case EmaiUIDNextCode:
					assert(inCommandData->sourceMailboxName != NULL);
					assert(folder != NULL);
					folder->SetNextUid(inResponse.responseText.param.number);
					break;
				case EmaiUIDValidityCode:
					assert(inCommandData->sourceMailboxName != NULL);
					assert(folder != NULL);
					folder->SetUidValidity(inResponse.responseText.param.number);
					break;
				case EmaiUnseenCode:
					assert(inCommandData->sourceMailboxName != NULL);
					assert(folder != NULL);
					folder->SetUnseen(inResponse.responseText.param.number);
					break;
				case EmaiReadOnlyCode:
					assert(inCommandData->sourceMailboxName != NULL);
					assert(folder != NULL);
					folder->SetReadOnly(true);
					break;
				case EmaiReadWriteCode:
					assert(inCommandData->sourceMailboxName != NULL);
					assert(folder != NULL);
					folder->SetReadOnly(false);
					break;
				case EmaiAppendUIDCode:
					assert(inCommandData->destMailboxName != NULL);
					assert(inResponse.responseText.param.uidPlus.sourceSequenceSet == NULL);
					assert(inResponse.responseText.param.uidPlus.destSequenceSet != NULL);
					folder = FindFolderByName(inCommandData->destMailboxName);
					//assume destination mailbox was listed previously
					assert(folder != NULL);
					folder->MessagesAppended(inCommandData->appendedMessageCount,
									inCommandData->appendedMessages,
									Emai::ImapSequenceSet(
							inResponse.responseText.param.uidPlus.destSequenceSet, true));
					break;
				case EmaiCopyUIDCode:
					assert(inCommandData->sourceMailboxName != NULL);
					assert(inCommandData->destMailboxName != NULL);
					assert(inResponse.responseText.param.uidPlus.sourceSequenceSet != NULL);
					assert(inResponse.responseText.param.uidPlus.destSequenceSet != NULL);
					assert(folder != NULL);
					std::cout << "Not implemented yet" << std::endl;
					break;
				case EmaiNoCode:
					break;
				case EmaiXUserDefinedCode:
					std::cout << "Unrecognized status response: ";
					if (inResponse.responseText.codeText != NULL)
						std::cout << inResponse.responseText.codeText << " ";
					if (inResponse.responseText.param.text != NULL)
						std::cout << inResponse.responseText.param.text;
					std::cout << std::endl;
					break;
				default:
					std::cout << "Unhandled status response: " <<
								inResponse.responseText.code << std::endl;
					break;
			}
			break;
		}
		case EmaiImapByeResponse:
			std::cout << "Disconnected" << std::endl;
			break;
		case EmaiImapBadResponse:
		case EmaiImapNoResponse:
			if ((inResponse.responseType & EmaiTaggedStatusResponse) != 0)
				std::cout << "Error: command failed with response: " <<
							inResponse.responseText.text << std::endl;
			else
				std::cout << "Warning: command failed with response: " <<
							inResponse.responseText.text << std::endl;
			break;
		default:
			std::cout << "Unhandled status response: " <<
						inResponse.responseText.code << std::endl;
			break;
	}
}

void
ImapResponseHandler::ProcessCapability(
		const EmaiCapabilityResponse& /*inResponse*/
	)
{
}

void
ImapResponseHandler::ProcessList(
		const EmaiListResponse& inResponse
	)
{
	
	std::auto_ptr<ImapFolder> folder(FindFolderByName(inResponse.mailboxName));
	if (folder.get() == NULL)
	{
		folder.reset(CreateFolderObject(inResponse));
		mFolders.push_back(folder.get());
	}
	else
		folder->SetAttributes(inResponse.attributes,
							Emai::StringList(inResponse.customAttributes, true));
	folder.release();
}

void
ImapResponseHandler::ProcessLsub(
		const EmaiLsubResponse& inResponse
	)
{
	std::auto_ptr<ImapFolder> folder(CreateFolderObject(inResponse));
	mSubscribedFolders.push_back(folder.get());
	folder.release();
}

void
ImapResponseHandler::ProcessMailboxStatus(
		const EmaiMailboxStatusResponse& inResponse
	)
{
	ImapFolder* folder = FindFolderByName(inResponse.mailboxName);
	assert(folder != NULL);
	folder->SetExists(inResponse.messages);
	folder->SetRecent(inResponse.recent);
	folder->SetNextUid(inResponse.uidNext);
	folder->SetUidValidity(inResponse.uidValidity);
	folder->SetUnseen(inResponse.unseen);
}

void
ImapResponseHandler::ProcessSearch(
		EmaiConstUniCharPtr inFolderName,
		const EmaiSearchResponse& inResponse
	)
{
	assert(inFolderName != NULL);
	ImapFolder* folder = FindFolderByName(inFolderName);
	assert(folder != NULL);

	folder->SetFoundMessageList(Emai::ImapSequenceSet(inResponse.messages, true));
}

void
ImapResponseHandler::ProcessFlags(
		EmaiConstUniCharPtr inFolderName,
		const EmaiFlagsResponse& inResponse
	)
{
	assert(inFolderName != NULL);
	ImapFolder* folder = FindFolderByName(inFolderName);
	assert(folder != NULL);

	folder->SetFlags(inResponse.flags,
					Emai::StringList(inResponse.customFlags, true));
}

void
ImapResponseHandler::ProcessExists(
		EmaiConstUniCharPtr inFolderName,
		const EmaiExistsResponse& inResponse
	)
{
	assert(inFolderName != NULL);
	ImapFolder* folder = FindFolderByName(inFolderName);
	assert(folder != NULL);

	folder->SetExists(inResponse.number);
}

void
ImapResponseHandler::ProcessRecent(
		EmaiConstUniCharPtr inFolderName,
		const EmaiRecentResponse& inResponse
	)
{
	assert(inFolderName != NULL);
	ImapFolder* folder = FindFolderByName(inFolderName);
	assert(folder != NULL);

	folder->SetRecent(inResponse.number);
}

void
ImapResponseHandler::ProcessExpunge(
		EmaiConstUniCharPtr inFolderName,
		const EmaiExpungeResponse& inResponse
	)
{
	assert(inFolderName != NULL);
	ImapFolder* folder = FindFolderByName(inFolderName);
	assert(folder != NULL);

	folder->MessageDeleted(inResponse.number);
}

void
ImapResponseHandler::ProcessFetch(
		EmaiConstUniCharPtr inFolderName,
		const EmaiFetchResponse& inResponse
	)
{
	assert(inFolderName != NULL);
	ImapFolder* folder = FindFolderByName(inFolderName);
	assert(folder != NULL);

	folder->AddMessageData(inResponse);
}

ImapFolder*
ImapResponseHandler::CreateFolderObject(
		const EmaiListResponse& inResponse
	) const
{
	return new ImapFolder(inResponse);
}

EmaiUint32
ImapResponseHandler::GetFoldersCount() const
{
	return static_cast<EmaiUint32>(mFolders.size());
}

ImapFolder*
ImapResponseHandler::GetFolderAtIndex(
		const EmaiUint32 inIndex
	) const
{
	assert(inIndex < GetFoldersCount());
	return mFolders[inIndex];
}

EmaiUint32
ImapResponseHandler::GetSubscribedFoldersCount() const
{
	return static_cast<EmaiUint32>(mSubscribedFolders.size());
}

ImapFolder*
ImapResponseHandler::GetSubscribedFolderAtIndex(
		const EmaiUint32 inIndex
	) const
{
	assert(inIndex < GetSubscribedFoldersCount());
	return mSubscribedFolders[inIndex];
}

